Require Import Utf8.
Require Import List.
Require Import IxFree.Base.

(* ========================================================================= *)
(* Implication *)

Definition I_arrow_func (P Q : IProp) : nat → Prop :=
  λ n, ∀ k, k ≤ n → proj1_sig P k → proj1_sig Q k.
Lemma I_arrow_func_monotone (P Q : IProp) :
  monotone (I_arrow_func P Q).
Proof.
unfold monotone; intros n H₁ k k_le_n; apply H₁; constructor; assumption.
Qed.

Definition I_arrow (P Q : IProp) : IProp.
exists (I_arrow_func P Q); apply I_arrow_func_monotone.
Defined.

Notation "A ⇒ B" := (I_arrow A B) (at level 90, right associativity).

Lemma I_arrow_intro {n : nat} {P Q : IProp} :
  (∀ k, k ≤ n → (k ⊨ P) → (k ⊨ Q)) → (n ⊨ (P ⇒ Q)).
Proof.
intro H; constructor; simpl.
unfold I_arrow_func; intros k Hle.
intro HP; apply Valid in HP.
specialize (H k Hle HP); destruct H; assumption.
Qed.

Lemma I_arrow_elim {n : nat} {P Q : IProp} :
  (n ⊨ P ⇒ Q) → (n ⊨ P) → (n ⊨ Q).
Proof.
intros Hf HP; destruct HP; destruct Hf as [ Hf ].
constructor; apply Hf; trivial.
Qed.

(* ========================================================================= *)
(* Conjunction *)

Definition I_conj_func (P Q : IProp) : nat → Prop :=
  λ n, proj1_sig P n ∧ proj1_sig Q n.
Lemma I_conj_func_monotone (P Q : IProp) :
  monotone (I_conj_func P Q).
Proof.
unfold monotone; destruct P; destruct Q; intros n;
  unfold I_conj_func; simpl; intuition.
Qed.

Definition I_conj (P Q : IProp) : IProp.
exists (I_conj_func P Q); apply I_conj_func_monotone.
Defined.

Notation "A '∧ᵢ' B" := (I_conj A B) (at level 80, right associativity).

Lemma I_conj_intro {n : nat} {P Q : IProp} :
  (n ⊨ P) → (n ⊨ Q) → (n ⊨ P ∧ᵢ Q).
Proof.
intros H₁ H₂; constructor; split.
+ destruct H₁; assumption.
+ destruct H₂; assumption.
Qed.

Lemma I_conj_elim1 {n : nat} {P Q : IProp} :
  (n ⊨ P ∧ᵢ Q) → (n ⊨ P).
Proof.
intro H; destruct H; destruct H; constructor; assumption.
Qed.

Lemma I_conj_elim2 {n : nat} {P Q : IProp} :
  (n ⊨ P ∧ᵢ Q) → (n ⊨ Q).
Proof.
intro H; destruct H; destruct H; constructor; assumption.
Qed.

(* ========================================================================= *)
(* IFF *)

Definition I_iff (P Q : IProp) := (P ⇒ Q) ∧ᵢ (Q ⇒ P).

Notation "A ⇔ B" := (I_iff A B) (at level 95).

Lemma I_iff_intro_M {n : nat} {P Q : IProp} :
  (∀ k, k ≤ n → ((k ⊨ P) <-> (k ⊨ Q))) → (n ⊨ P ⇔ Q).
Proof.
intro H; apply I_conj_intro; apply I_arrow_intro; apply H.
Qed.

Lemma I_iff_elim_M {n : nat} {P Q : IProp} :
  (n ⊨ P ⇔ Q) → ((n ⊨ P) <-> (n ⊨ Q)).
Proof.
intro H; split; intro H'.
+ apply I_conj_elim1 in H; apply (I_arrow_elim H H').
+ apply I_conj_elim2 in H; apply (I_arrow_elim H H').
Qed.

(* ========================================================================= *)
(* Quantifiers *)

Definition I_forall_func (A : Type) (P : A → IProp) : nat → Prop :=
  λ n, ∀ x : A, proj1_sig (P x) n.
Lemma I_forall_func_monotone (A : Type) (P : A → IProp) :
  monotone (I_forall_func A P).
Proof.
unfold monotone; unfold I_forall_func; intros n H₁ x; specialize H₁ with x.
destruct (P x); simpl; simpl in H₁; auto.
Qed.

Definition I_forall (A : Type) (P : A → IProp) : IProp.
exists (I_forall_func A P); apply I_forall_func_monotone.
Defined.

Notation "'∀ᵢ' x .. y , P" :=
  (I_forall _ (fun x => .. (I_forall _ (fun y => P)) .. ))
  (at level 200, x binder, y binder, right associativity).

Definition I_exists_func (A : Type) (P : A → IProp) : nat → Prop :=
  λ n, ∃ x : A, proj1_sig (P x) n.
Lemma I_exists_func_monotone (A : Type) (P : A → IProp) :
  monotone (I_exists_func A P).
Proof.
unfold monotone; unfold I_exists_func; intros n H₁; 
  destruct H₁ as [x H₁]; exists x.
destruct (P x); simpl; auto.
Qed.

Definition I_exists (A : Type) (P : A → IProp) : IProp.
exists (I_exists_func A P); apply I_exists_func_monotone.
Defined.

Notation "'∃ᵢ' x .. y , P" :=
  (I_exists _ (fun x => .. (I_exists _ (fun y => P)) .. ))
  (at level 200, x binder, y binder, right associativity).

Lemma I_forall_intro {n : nat} {A : Type} {P : A → IProp} :
  (∀ x : A, (n ⊨ P x)) → (n ⊨ ∀ᵢ x, P x).
Proof.
intro H; constructor; intro x.
destruct (H x); auto.
Qed.

Lemma I_forall_elim {n : nat} {A : Type} {P : A → IProp} :
  (n ⊨ ∀ᵢ x : A, P x) → ∀ x, n ⊨ P x.
Proof.
intros H x; constructor; destruct H; apply H.
Qed.

(* ========================================================================= *)
(* Later *)

Definition I_later_func (P : IProp) : nat → Prop :=
  λ n, 
  match n with
  | 0   => True
  | S n => proj1_sig P n
  end.
Lemma I_later_func_monotone (P : IProp) :
  monotone (I_later_func P).
Proof.
unfold monotone; destruct P; unfold I_later_func; simpl; intros n H₁.
destruct n; auto.
Qed.

Definition I_later (P : IProp) : IProp.
exists (I_later_func P); apply I_later_func_monotone.
Defined.

Notation "▷ P" := (I_later P) (at level 30).

Fixpoint IRel_later (l : list Type) : IRel l → IRel l :=
  match l return IRel l → IRel l with
  | nil    => λ R, ▷R
  | A :: l => λ R x, IRel_later l (R x)
  end.

(* ========================================================================= *)
(* Structural rules for ▷ *)

Lemma I_later_zero (P : IProp) : 0 ⊨ ▷P.
Proof.
repeat constructor.
Qed.

Lemma I_later_shift {n : nat} {P : IProp} :
  (S n ⊨ ▷P) <-> (n ⊨ P).
Proof.
split; intro H; destruct H; constructor; apply H.
Qed.

Lemma I_later_arrow_down {n : nat} {P Q : IProp} :
  (n ⊨ ▷(P ⇒ Q)) → (n ⊨ ▷P ⇒ ▷Q).
Proof.
intro H; apply I_arrow_intro; intros k Hle.
apply (I_valid_monotone _ k n Hle) in H.
intro HP; destruct k.
+ apply I_later_zero.
+ apply I_later_shift; 
  apply (proj1 I_later_shift) in H; apply (proj1 I_later_shift) in HP.
  eapply I_arrow_elim; eassumption.
Qed.

Lemma I_later_forall_up {n : nat} {A : Type} {P : A → IProp} :
  (n ⊨ ∀ᵢ x, ▷ P x) → (n ⊨ ▷ ∀ᵢ x, P x).
Proof.
intro H; destruct n; [ apply I_later_zero | apply I_later_shift ].
apply I_forall_intro; intro x.
eapply I_forall_elim in H.
apply I_later_shift; eassumption.
Qed.

(* ========================================================================= *)
(* Main rules for ▹ *)

Lemma I_later_intro {n : nat} (P : IProp) :
  (n ⊨ P) → (n ⊨ ▷ P).
Proof.
intro H; destruct n; [ apply I_later_zero | apply I_later_shift ].
apply I_valid_monotone_S; assumption.
Qed.

Lemma I_loeb_induction {n : nat} (P : IProp) :
  (∀ k, k ≤ n → (k ⊨ ▷P) → (k ⊨ P)) → (n ⊨ P).
Proof.
induction n; intro H; apply H; trivial.
+ apply I_later_zero.
+ apply I_later_shift.
  apply IHn; intros k Hle; apply H.
  constructor; assumption.
Qed.

(* ========================================================================= *)
(* Tactics *)

Ltac iintro_named X :=
  match goal with
  | [ |- ?N ⊨ ?P ⇒ ?Q ] =>
    apply (@I_arrow_intro N P Q);
    let K := fresh "K" in
    let L := fresh "L" in
    intros K L X;
    repeat 
      match goal with
      | [ A : N ⊨ ?R |- _ ] => 
        apply (I_valid_monotone R K N L) in A
      end;
    clear L;
    clear N;
    rename K into N
  | [ |- ?N ⊨ I_forall ?A ?P ] =>
    apply (@I_forall_intro N A P); intro X
  end.

Ltac iintro_anon :=
 match goal with
  | [ |- ?N ⊨ ?P ⇒ ?Q ] =>
    apply (@I_arrow_intro N P Q);
    let H := fresh "H" in
    let K := fresh "K" in
    let L := fresh "L" in
    intros K L H;
    repeat 
      match goal with
      | [ A : N ⊨ ?R |- _ ] => 
        apply (I_valid_monotone R K N L) in A
      end;
    clear L;
    clear N;
    rename K into N
  | [ |- ?N ⊨ I_forall ?A ?P ] =>
    apply (@I_forall_intro N A P); intro
  | [ |- ?N ⊨ ▷ ?P ] =>
    apply (I_later_intro P)
  | [ |- ?N ⊨ ( ?P )ᵢ ] =>
    apply (I_Prop_intro P)
  end.

Tactic Notation "iintro" ident(H) := iintro_named H.
Tactic Notation "iintro" := iintro_anon.

Ltac iapply H :=
  first [ apply (I_arrow_elim H) | apply (I_forall_elim H) ].

Ltac ispecialize_arrow H :=
  let T := type of H in
  match T with
  | (?N ⊨ ?P ⇒ ?Q) =>
    let J := fresh in
    assert (J : N ⊨ Q); [ iapply H | clear H; rename J into H ]
  end.

Ltac ispecialize_forall H X :=
  let T := type of H in
  match T with
  | (?N ⊨ I_forall ?A ?P) =>
    let J := fresh in
    assert (J : N ⊨ P X); 
    [ apply (I_forall_elim H X) 
    | clear H; rename J into H; try (cbv beta in H) 
    ]
  end.

Tactic Notation "ispecialize" hyp(H) := ispecialize_arrow H.
Tactic Notation "ispecialize" hyp(H) constr(X) := ispecialize_forall H X.

Ltac igeneralize H :=
  let T := type of H in
  match T with
  | (?N ⊨ ?P) =>
    match goal with
    | [ |- (N ⊨ ?Q) ] =>
      let J := fresh in
      assert (N ⊨ P ⇒ Q); [ idtac | apply (I_arrow_elim J H) ]
    end
  end.

Ltac isplit := apply I_conj_intro.

Ltac later_up := 
  match goal with
  | [ |- _ ⊨ _ ⇒ _ ] => apply I_later_arrow_down
  end.

Ltac later_down :=
  match goal with
  | [ |- _ ⊨ ▷ I_forall _ _ ] => apply I_later_forall_up
  end.

Ltac loeb_induction :=
  match goal with
  | [ |- ?N ⊨ ?P ] =>
    apply (@I_loeb_induction N P);
    let K  := fresh "K" in
    let L  := fresh "L" in
    let IH := fresh "IH" in
    intros K L IH;
    repeat 
      match goal with
      | [ A : N ⊨ ?R |- _ ] => 
        apply (I_valid_monotone R K N L) in A
      end;
    clear L;
    clear N;
    rename K into N
  end.

Ltac later_shift :=
  match goal with
  | [ |- ?N ⊨ ▷ _ ] =>
    destruct N as [ | N ]; [ apply I_later_zero | apply I_later_shift ];
    repeat match goal with
    | [ H : S N ⊨ ▷ _ |- _ ] => apply (proj1 I_later_shift) in H
    | [ H : S N ⊨ _ |- _ ] => apply I_valid_monotone_S in H
    end
  end.